<?php
/**
 * Plugin Name: Subtotal
 * Plugin URI: https://subtotal.ru/support/integraczii/integracziya-s-internet-magazinom-woocommerce
 * Description: Subtotal integration with WooCommerce
 * Author: Subtotal
 * Author URI: https://subtotal.ru
 */
require_once ABSPATH . 'wp-admin/includes/media.php';

/** Типы товаров */
// услуги
define('GOOD_TYPE_ID_SERVICE', 1);
// товары
define('GOOD_TYPE_ID_GOODS', 2);
// Производимый товар (тех карта, производимая заранее)
define('GOOD_TYPE_ID_COMPOSED', 3);
// Производимый на кассе (тех карта, производимая на кассе)
define('GOOD_TYPE_ID_COMPOSED_CASHBOX', 4);
// Одноразовый товар (тех карта, произведённая по шаблону)
define('GOOD_TYPE_ID_COMPOSED_TEMPLATE', 5);
// Модификации
define('GOOD_TYPE_ID_MODIFICATION', 6);

/**
 * Проверка блокировки обновления
 * @return bool true если пересечения есть, false если нет
 */
function st_hasBlock(): bool {
	$block = get_option('subtotal_block');
	$block_ttl = get_option('subtotal_block_ttl');

	if (
		isset($block)
		&& is_array($block)
		&& isset($block['timestamp'])
		&& (int) $block['timestamp'] == $block['timestamp']
		&& $block['timestamp'] * $block_ttl < time()
	)
	{
		return true;
	}

	return false;
}

/**
 * Обновление блокировки
 */
function st_updateBlock() {
	$block = get_option('subtotal_block');

	if ( !isset($block) || !is_array($block) ) {
		$block = Array();
	}

	$block['timestamp'] = time();
	$block['blocked'] = true;
	update_option('subtotal_block', $block);
}

/**
 * Основная процедура импорта.
 * В случае успеха возвращает true.
 * Иначе null, а должна причину неудачи.
 *
 * @param array $good Данные товара в формате Subtotal
 * @param bool $auto Признак импорта по расписанию
 *
 * @return id товара или -1 если что-то пошло не так
 */
function st_insertOrUpdateProduct($good, $auto = false): int
{
	global $SYNCHRONIZED_GOODS_IDS;

	if (st_hasBlock()) {
		st_log_info('Импорт товаров в WooCommerce не выполнен из-за блокировки');
		return -1;
	}

	st_updateBlock();

	/**
	 * В текущей реализации отсутствует кеширование,
	 * поэтому импорт не быстрый.
	 */

	/**
	 * Прилетел товар.
	 * Это может быть обычный товар, или модификация.
	 * $good['id'] tag cache
	 */
	$good_id = $good['id'];

	if (in_array($good_id, $SYNCHRONIZED_GOODS_IDS)) {
		return $good_id;
	}

	$log_message = 'Execution: st_insertOrUpdateProduct, '
					. 'product_id: ' . $good_id
					. ', auto: ' . $auto;
	st_log_info($log_message);

	// Проверяем что у товара хотя бы один тег из списка тегов, указанных пользователем
	$import_tags = get_option('subtotal_import_good_tags');
	if (st_checkProductInTags($good, $import_tags) == false) {
		$log_message = 'Product ' . $good_id
						. ' has no tags: '
						. implode(', ', $import_tags);
		st_log_notice($log_message);
		return -1;
	}

	/**
	 * Детальная информация о продукте
	 * Если $good - модификация, то `$good_details['good']['modifications']`
	 * будет содержать информацию о модификациях.
	 */
	$good_details = st_getGoodDetails($good_id);
	if (!$good_details) {
		$log_message = 'No product details returned for product_id: ' . $good_id;
		st_log_notice($log_message);
		return -1;
	}

	// id базового продукта будет отличаться для вариаций
	$base_good_id = $good_details['good']['id'];
	if (!$base_good_id) {
		$log_message = 'No product details for product_id: ' . $good_id;
		st_log_debug($log_message);
		return -1;
	}

	// Опции импорта
	$import_photos = get_option('subtotal_import_photos');
	$update_stocks = get_option('subtotal_update_stocks');
	$import_description = !is_null(get_option('subtotal_import_description')) ? get_option('subtotal_import_description') : '';
	$import_tags = !is_null(get_option('subtotal_import_tags')) ? get_option('subtotal_import_tags') : '';

	// Ищем этот продукт среди уже загруженных по метаданным
	// todo: а добавляется ли мета при добавлении товаров?
	$product_id = wp_getPostIdByMetaData('_subtotal_id', $base_good_id);
	$subtotal_data = $good;

	// Вытащить товар из корзины, если он там
	if (!$product_id) {
		// TODO: протестировать
		wp_untrash_post($product_id);
	}

	// Обработка обычного товара
	if ($good['good_type_id'] !== GOOD_TYPE_ID_MODIFICATION) {
		$log_message = 'Process regular product "' . $good_id . '"';
		st_log_debug($log_message);

		$good_details['good']['avial'] = $good['avial'];
		// TODO: нужно кешировать, но при этом учитывать что продукт - 'simple'
		$product = process_base_product($product_id, $good_details, $good, 'simple');
	} else {
		if (empty($good_details['good']['modifications'])) {
			throw new Exception('todo: нет модификаций для товара и т.п.');
		}

		/**
		 * Подготовка данных
		 * Индекс для поиска имени модификации
		 */
		$mod_name_idx = [];
		foreach ($good_details['good']['properties'] as $prop) {
			$key = $prop['id'];
			$name = $prop['name'];
			$mod_name_idx[$key] = $name;
		}

		// TODO: нужно кешировать, но при этом учитывать что продукт - 'variable'
		$product = process_base_product($product_id, $good_details, $good, 'variable');
		$product_id = $product->get_id();
		add_custom_attrs($product, $good_details['good']['properties']);

		foreach ($good_details['good']['modifications'] as $modification) {
			$new_variation_attrs = [];
			foreach ($modification['properties'] as $mod_prop) {
				$attr_name = $mod_name_idx[$mod_prop['prop_id']];
				/**
				 * Рассадник багов и говнокода:
				 */

				// 1. В названии атрибута тоже может быть изменён регистр
				$attr_name = mb_strtolower($attr_name);
				// 2. urlencode возвращает в верхнем регистре, а в БД атрибут храниться в нижнем.
				$attr_name = mb_strtolower(urlencode($attr_name));
				// 3. Скорее всего есть штатные методы подготовки массива атрибутов
				$attr_name = 'attribute_' . $attr_name;

				$attr_value = $mod_prop['value'];

				/**
				 * В Subtotal есть бага: может прилететь пустая модификация.
				 * TODO: пофиксить
				 */
				if ($attr_value) {
					$new_variation_attrs[$attr_name] = $attr_value;
				}
			}

			process_variation_product(
				$product_id,
				$modification,
				$good,
				$new_variation_attrs
			);
			// К продукту добавилась вариация. Следует его перечитать.
			$product = load_product($product_id, $base_good_id, 'variable');
		}
	}

	$SYNCHRONIZED_GOODS_IDS[] = $good_id;
	return $good_id;
}

/**
 * Загрузка товара из WC.
 * При вызове процедуры с недопустимым аргументом выбрасывается исключение.
 * Предполагается, что проверки аргумента на допустимость осуществляются
 * до вызова процедуры.
 */
function load_product(
	int $product_id,
	int $good_id,
	?string $product_type = ''
): WC_Product {
	global $LOADED_PRODUCTS;

	$cache_key = $product_id . $product_type;
	if (isset($LOADED_PRODUCTS[$cache_key])) {
		return $LOADED_PRODUCTS[$cache_key];
	}

	if (!$product_id || $product_id < 1) {
		// Автор, проверь аргумент до вызова процедуры!
		$log_message = 'Unacceptable product id: ' . $product_id;
		throw new Error($log_message);
	}

	$log_message = 'Try to load Subtotal good: ' . $good_id
					. ' as WooCommerce product: ' . $product_id . '. ';

	if ($product_type == 'variable') {
		$log_message .= 'Instantiate Variable product';
		$product = new WC_Product_Variable($product_id);
	} elseif ($product_type == 'external') {
		$log_message .= 'Instantiate External product';
		$product = new WC_Product_External($product_id);
	} elseif ($product_type == 'grouped') {
		$log_message .= 'Instantiate Grouped product';
		$product = new WC_Product_Grouped($product_id);
	} else {
		$log_message .= 'Simple product';
		$product = new WC_Product($product_id);
	}

	$log_message .= ' load successfully';
	st_log_debug($log_message);

	$LOADED_PRODUCTS[$cache_key] = $product;
	return $product;
}

/**
 * Обработка обычного товара (не модификации).
 * TODO: $product_type - в php нет enum'ов. Задокументировать допустимые значения
 * и проверять их на входе (тогда, имхо, документировать не надо :)
 */
function process_base_product(
	?int $product_id,
	array $good_details,
	array $good_details_ex,
	string $product_type
) {
	global $BASE_PRODUCTS;

	$good = $good_details['good'];
	$good_id = $good['id'];
	$good_name = $good['name'];

	if (isset($BASE_PRODUCTS[$good_id][$product_type])) {
		return $BASE_PRODUCTS[$good_id][$product_type];
	}

	$log_message = 'Try to create new product for good "' . $good_name . '" (id: ' . $good_id . ')';
	st_log_debug($log_message);

	$price_id = get_option('subtotal_import_price');
	$price_value = st_getProductPriceValue($good, $price_id);

	/**
	 * Что может пойти не так:
	 * 1. В WC продукт обычный, из Subtotal он прилетел как модификация.
	 * 2. Или наоборот.
	 * Создан кейс для тестирования. Исправления по его результатам.
	 */
	if ($product_id) {
		$product = load_product($product_id, $good['id'], $product_type);
	} elseif ($product_type == 'variable') {
		$product = new WC_Product_Variable();
	} else {
		$product = new WC_Product_Simple();
	}

	fill_product_data($product, $good, $good_details_ex);

	/**
	 * 1. Не оптимально - можно заранее получить id категорий и не запрашивать их каждый раз
	 * 2. Логика ниже не учитывает был ли товар добавлен в категорию на сайте:
	 * на сайте товар добавлен в более иные группы, чем в Subtotal.
	 * Например, на сайте кофе добавлен в группы "напитки" и "2go",
	 * а в Subtotal в группы "напитки" и "безалкагольные".
	 * После синхронизации будет как в Subtotal - "напитки" и "безалкагольные".
	 */
	$product_categories = [];
	foreach ($good_details_ex['tags'] as $tag) {
		$existed_term = term_exists($tag, 'product_cat');
		if ($existed_term) {
			$product_categories[] = $existed_term['term_id']; // или term_taxonomy_id?
		}
	}
	if ($product_categories) {
		$product->set_category_ids($product_categories);
	}

	$product_id = $product->save();
	$product_id = intval($product_id);

	update_images($product, $good);
	update_post_meta($product_id, '_subtotal_id', $good['id']);
	// Сохранять $good или полностью $good_details - вот в чём вопрос
	update_post_meta($product_id, '_subtotal_data', $good);
	update_post_meta($product_id, '_subtotal_updated', time());

	$BASE_PRODUCTS[$good_id][$product_type] = $product;
	return $product;
}

/**
 * Заполнение товара данными.
 * Мутатор.
 */
function fill_product_data(
	object &$product,
	array $good,
	array $good_details_ex,
	?int $parent_id = 0,
	array $variation_attrs = []
) {
	// Выяснение цены на продукт по выбранному прайсу
	$price_id = get_option('subtotal_import_price');
	$price_value = st_getProductPriceValue($good, $price_id);

	// Структура с данными, которыми обновится продукт.
	$product_data = [
		'name' => $good['name'],
		'slug' => $good['name'],
		'status' => 'publish',
		'catalog_visibility' => 'visible',
		'sku' => $good['article'],
		'regular_price' => $price_value,
		//'sale_price' =>  '',
		'tax_status' => 'none',
		//'tax_class' => '',
		//'manage_stock' => true, // Управляем этим в зависимости от настройки
		'stock_status' => 'instock',
		'sold_individually' => false,
		'parent_id' => $parent_id,
		'reviews_allowed' => true,
		//'default_attributes' => [],
		'attributes' => $variation_attrs,
		//'category_ids' => [],
	];

	// Обновлять остатки только если стоит галка
	if (get_option('subtotal_update_stocks')) {
		$product_data['manage_stock'] = true;
		$product_data['stock_quantity'] = $good_details_ex['avial']; // да, avial
	}

	if (!empty($good['description'])) {
		$product_data['description'] = $good['description'];
	}

	foreach ($product_data as $function => $value) {
		$product->{'set_' . $function}($value);
	}
}

/**
 * Обновление товара-модификации
 */
function process_variation_product(
	int $parent_product_id,
	array $good_details, // TODO: описать чем этот аргумент отличается от $good_details_ex
	array $good_details_ex,
	array $mod_attrs
) {
	$good_id = $good_details['id'];
	$log_message = 'Try to create new variation for Subtotal good: ' . $good_id
					. '(' . $good_details['name'] . ')'
					. ', parent WooCommerce product: ' . $parent_product_id;

	st_log_debug($log_message);

	if (isset($VARIATION_PRODUCTS[$good_id]['product'])) {
		return;
	}

	$variation = new WC_Product_Variation();

	fill_product_data(
		$variation,
		$good_details,
		$good_details_ex,
		$parent_id = $parent_product_id,
		$variation_attrs = $mod_attrs
	);

	$variation_id = $variation->save();

	// Импорт только главного фото без галереи (если включен)
	update_images($variation, $good_details, true);

	$VARIATION_PRODUCTS[$good_id] = ['product' => $variation];

	update_post_meta($variation_id, '_subtotal_data', $good_details_ex);
	update_post_meta($variation_id, '_subtotal_updated', time());
}

/**
 * Импорт главного фото и галереи, если включена соответствующая настройка.
 * Если передан true аргумент $force, то импорт фото будет осуществлён
 * независимо от состояния настройки импорта фото.
 */
function update_images(
	WC_Product $product,
	array $good,
	bool $only_main_image = false,
	?bool $force = false
) {
	$import_photos = get_option('subtotal_import_photos');
	if (!$import_photos && !$force) {
		return;
	}

	$product_type = $product->get_type();

	/**
	 * Может обрабатываться как товар, так и его модификация.
	 * У этих сущностей разные поля для хранения изображений.
	 * В случае товара - это поля imag1,.., image4
	 * В случае модификации - это поле image.
	 */

	$images_url = [];
	if (isset($good['image'])) {
		// Модификация
		if (isset($good['image']['url'])) {
			array_push($images_url, $good['image']['url']);
		}
	} else {
		// Товар
		$image_range = [1, 2, 3, 4];
		foreach ($image_range as $image_idx) {
			$image_name = 'image' . $image_idx;
			if (isset($good[$image_name]['url'])) {
				array_push($images_url, $good[$image_name]['url']);
			}
		}
	}

	if (!count($images_url)) {
		return;
	}

	$product_id = $product->get_id();
	$photo_data = st_getProductPhoto($good, $images_url, $product_id);
	if ($photo_data['main_image']) {
		// Как правильно добавлять картинку к товару - загадка.
		// Строка ниже - вроде бы ок, но картинка не отображается.
		$product->set_image_id($photo_data['main_image']);
		// Поэтому картинка добавляется через метадату
		update_post_meta(
			$product_id,
			'_thumbnail_id',
			$photo_data['main_image']
		);
	}

	if ($only_main_image) {
		return;
	}

	if (count($photo_data['gallery']) > 0) {
		// Как правильно добавлять галерею к товару - загадка.
		// Строка ниже - вроде бы ок, но галерея не отображается.
		$product->set_gallery_image_ids($photo_data['gallery']);
		// Поэтому галерея добавляется через метадату
		update_post_meta(
			$product_id,
			'_product_image_gallery',
			implode(',', $photo_data['gallery'])
		);
	}
}

/**
 * Свойства Subtotal в атрибуты WC
 */
function props_to_attributes(array $props): array
{
	$result = [];
	foreach ($props as $property) {
		$name = $property['name'];
		$options = [];
		foreach ($property['values'] as $prop_val) {
			$options[] = $prop_val['value'];
		}
		$result[$name] = $options;
	}
	return $result;
}

/**
 * Создание кастомных атрибутов у базового товара
 *
 * $props - массив вида:
 * [
 * 	{
 *     	id: ,
 *     	name: ,
 *     	values: [{pv_id: , value:}, ...]
 * 	},
 * ...
 * ]
 */
function add_custom_attrs($product, $props)
{
	global $PRODUCT_VARIATIONS;

	$product_type = $product->get_type();
	$product_id = $product->get_id();

	if ($product_type !== 'variable') {
		$log_message = 'Добавлять атрибуты можно только к товару с типом "variable"'
		. ', тип обрабатываемого товара: ' . $product_type
		. ', id товара: ' . $product_id
		. ', наименование товара: ' . $product->get_name();
		throw new Exception($log_message);
	}

	if (in_array($product_id, $PRODUCT_VARIATIONS)) {
		return;
	}

	$wp_attrs = $product->get_variation_attributes();
	$st_attrs = props_to_attributes($props);

	/**
	 * Логика добавления атрибута проста:
	 * на основании принятых данных создаются новые атрибуты.
	 * Старые атрибуты затираются.
	 */
	foreach ($st_attrs as $key => $value) {
		$attr = new WC_Product_Attribute();
		$attr->set_name($key);
		// 0 означает что атрибут локальный
		$attr->set_id(0);
		// $attr->set_options(array_unique($value));
		$attr->set_options($value);
		$attr->set_visible(true);
		$attr->set_variation(true);
		$new_attrs[] = $attr;
	}

	$product->set_attributes($new_attrs);
	$product->save();
	$PRODUCT_VARIATIONS[] = $product_id;
}

/**
 * Экспортирует продажу из WooCommerce в Subtotal
 * Используется как хук при обновлении статуса заказа (т.е. его подтверждении).
 * См. add_action('woocommerce_update_order', 'subtotal_export');
 *
 * @param int $order_id Идентификатор продажи WooCommerce
 */
function subtotal_export($order_id)
{
	/**
	 * Статическая переменная введена для того,
	 * чтобы предотвратить повторный вызов этой процедуры.
	 * Чего греха таить - костыль!
	 */
	static $export_fired = false;

	if ($export_fired) {
		return;
	}

	$export_fired = true;

	$access = get_option('subtotal_access');
	if (!$access) {
		// todo: save log
		return;
	}

	$pos_id = get_option('subtotal_import_pos');
	if (!$pos_id) {
		// todo: save log
		return;
	}

	$stores = (array) get_option('subtotal_import_stores');
	if (!count($stores)) {
		// todo: save log
		return;
	}

	/**
	 * TODO: пока не реализована функциональность выборки остатков
	 * по нескольким складам, склады хранятся как массив с одним элементом.
	 * tag: multi-stores
	 */
	$store_id = $stores[0];

	$price_id = get_option('subtotal_import_price');
	if (!$price_id) {
		// todo: save log
		return;
	}

	$order = wc_get_order($order_id);
	$order_data = $order->get_data();

	// Найти или создать пользователя, сначала по e-mail
	$partner_found = false;
	$partner_id = null;
	$partner_email = $order_data['billing']['email'];
	$partner_phone = $order_data['billing']['phone'];
	$address = get_post_meta($order_id, '_billing_address_index')[0];
	$partners_exists = st_filterPartners($partner_email);

	if (is_array($partners_exists->data)) {
		if (
			$partner_email != ''
			&& array_search($partner_email, array_column($partners_exists->data, 'email')) !== false
		) {
			$partner_found = true;
			$partner_id = $partners_exists->data['id'];
		} else if (
			$partner_phone != ''
			&& array_search($partner_email, array_column($partners_exists->data, 'phone')) !== false
		) {
			$partner_found = true;
			$partner_id = $partners_exists->data['id'];
		}
	}

	// Если клиента нет - создаём нового
	if (!$partner_found) {
		// Только если заполнено хоть какое-то поле
		if (
			$order_data['billing']['first_name']
			|| $order_data['billing']['last_name']
			|| $partner_email
			|| $partner_phone
		) {
			$partner_data = [
				'contacts' => [[
					'fname' => $order_data['billing']['first_name'],
					'lname' => $order_data['billing']['last_name'],
					'mname' => '',
					'email' => $partner_email,
					'phone' => $partner_phone,
					'addr_fact' => $address,
				]],
				'partner' => [
					'legal_type_id' => 3,
					'partner_type_id' => 5,
					'email' => $partner_email,
					'phone' => $partner_phone,
					'description' => 'Создан автоматически к заказу #' . $order_data['id'],
				]
			];
			$new_partner_data = st_createPartner($partner_data);
			$partner_id = $new_partner_data['id'];
		} else {
			// писать лог, что партнёр не найден
		}
	}

	// Комментарий к заказу
	$comment = 'WooCommerce';
	// Форматировать дату на основе настроек WordPress'а
	// $comment .= str_replace('T', ' ', $order_data['date_created']);
	$comment .= ' создан заказ #';
	$comment .= $order_data['id'];
	$comment .= ' на сайте ' . get_site_url();

	if (!empty($order_data['customer_note'])) {
		$comment .= '\nКомментарий покупателя: ' . $order_data['customer_note'];
	}

	if (!empty($address)) {
		$comment .= '\nАдрес: $address';
	}

	$exists_sale_id = null;
	if (get_post_meta($order_id, '_subtotal_sale_id')) {
		$exists_sale_id = get_post_meta($order_id, '_subtotal_sale_id')[0];
		$sale_uid = get_post_meta($order_id, '_subtotal_sale_uid')[0];
	}

	if (empty($sale_uid)) {
		$sale_uid = uniqid();
	}

	$sale = [
		'goods' => [],
		'outcome' => [
			'pos_id' => $pos_id,
			'store_id' => $store_id,
			'description' => $comment,
			'partner_id' => $partner_id,
		],
		'price' => $price_id,
		'uid' => $sale_uid,
	];

	// Заполнить продукты
	foreach ($order->get_items() as $item) {
		$item_data = $item->get_data();
		$product_id = $item_data['product_id'];
		$variation_id = $item_data['variation_id'];
		// Если покупают вариацию, то id товара надо брать из вариации
		if ($variation_id) {
			$subtotal_product_data = get_post_meta($variation_id, '_subtotal_data')[0];
			$good = [
				'id' => $subtotal_product_data['id'],
			];
		}
		// Если не вариация - то id основного товара
		else {
			$subtotal_product_data = get_post_meta($product_id, '_subtotal_data')[0];
			$good = [
				'id' => $subtotal_product_data['id'],
			];
		}

		$sale['goods'][] = [
			'id' => null,
			'cnt' => $item_data['quantity'],
			'good' => $good,
			'price' => $item_data['total'],
		];
	}

	$result = st_exportSale($sale, $exists_sale_id);
	if (isset($result['status']) && !$result['status']) {
		// По неведомой причине брошенное исключение не пробрасывается клиенту.
		// Другими словами, если при оформлении заказа возникло исключение,
		// то пока не ясно что делать: или пробросить его пользователю,
		// и/или дёрнуть webhook, чтобы сотрудники могли оперативно обработать
		// заказ с ошибкой.
		// throw new Exception($result['error']);
		st_log_info('Ошибка создания заказа: ' . $result['error']);
	}

	if ($result['id']) {
		$sale_id = $result['id'];
		update_post_meta($order_id, '_subtotal_sale_id', $sale_id);
		update_post_meta($order_id, '_subtotal_sale_uid', $sale_uid);

		$sale_data = st_getSale($sale_id);
		$pay_status_id = $sale_data['pay_status_id'];

		if (in_array($order_data['status'], ['processing', 'completed']) && $pay_status_id != 100) {
			// Если оплата наличными курьеру - это наличные, остальное считается картой
			if ($order_data['payment_method'] == 'cod') {
				$pay_type = 1; // Наличные
			} else {
				$pay_type = 2; // Карта
			}

			$payData = [
				'description' => 'Заказ #' . $order_data['id'] . ' помечен как оплаченный (' . $order_data['status'] . ') способом ' . $order_data['payment_method_title'],
				'no_shift' => 0,
				// TODO: формат даты из настроек WordPress'а
				'pay_date' => date('Y-m-d') . 'T' . date('H:i:s.v') . 'Z', // Пример 2020-05-13T13:50:51.598Z
				'pay_sum' => $order_data['total'],
				'pay_type' => $pay_type,
				'pos' => $pos_id,
			];

			$result = st_payOrder($sale_id, $payData);
		}
	}
}

/**
 * Устанавливает фото продукта (название ф-и не соответствует поведению)
 *
 * @param array $good Данные о продукте (чтобы подставить название)
 * @param array $images_url Массив URL с фотографиями
 */
function st_getProductPhoto(array $good, array $images_url): array
{
	$result = [
		'main_image' => null,
		'gallery' => [],
	];

	if (!get_option('subtotal_import_photos')) {
		return $result;
	}

	$good_name = $good['name'];
	$main_image_set = false;

	if (count($images_url) > 0) {
		foreach ($images_url as $image_url) {
			$id = fetchAndPostImage($good_name, $image_url);
			if ($id == -1) {
				continue;
			}

			if (!$main_image_set) {
				$result['main_image'] = $id;
				$main_image_set = true;
			} else {
				$result['gallery'][] = $id;
			}
		}
	}

	return $result;
}

/**
 * Скачивает картинку из Subtotal, загружает её в WP и возвращает её id
 *
 * @param string $good_name Название продукта будет установлено как название картинки
 * @param string $image_url Ссылка на картинку
 *
 * @return int Идентификатор картинки в WordPress в случае успеха. Иначе: -1
 */
function fetchAndPostImage(string $good_name, string $image_url): int
{
	if ($image_url != '') {
		// Ссылка может меняться, а вот хэш картинки - нет
		$image_data = st_getImage($image_url);
		if (!isset($image_data)) {
			return -1;
		}

		$hash = hash('sha256', $image_data['data']);
		$id = wp_getPostIdByMetaData('_subtotal_image_hash', $hash);

		if ($id) {
			$file_path = get_attached_file($id);
			if (file_exists($file_path)) {
				return $id;
			}
		}

		// Картинки доступны только авторизованному пользователю
		// sideload нельзя передать куку авторизации
		// Поэтому загружаем файл в папку загрузки
		// и передаём ссылку как файл для скачивания
		$wp_upload_dir = wp_upload_dir();
		$tmp_file_path = $wp_upload_dir['path'] . $image_data['name'];
		file_put_contents($tmp_file_path, $image_data['data']);
		// Можно попробовать писать в tmpfile(),
		// тогда удалять вручную не потребуется 8)
		$tmp_file_url = $wp_upload_dir['url'] . $image_data['name'];
		$id = media_sideload_image($tmp_file_url, 0, $good_name, 'id');
		$id = intval($id);
		update_post_meta($id, '_subtotal_image_hash', $hash);
		unlink($tmp_file_path);

		return $id;
	}

	return -1;
}

/**
 * Получить из WordPress id поста по ключу и значению из postmeta
 *
 * @param string $key Ключ
 * @param mixed $value Значение
 *
 * @return int Идентификатор поста WordPress к которому привязаны этот ключ и значение
 */
function wp_getPostIdByMetaData(string $meta_key, $value)
{
	global $wpdb;
	$post_id = null;

	$query = $wpdb->prepare(
		'SELECT post_id FROM ' . $wpdb->postmeta . ' WHERE meta_key = %s AND meta_value = %s',
		$meta_key,
		$value
	);

	$data = $wpdb->get_row($query, ARRAY_A);
	if (isset($data) && isset($data['post_id'])) {
		$post_id = intval($data['post_id']);
	}

	/**
	 * Костылёк: если в БД остаются следы от предыдущих товаров.
	 * Некостыльное решение - при удалении товара чистить БД.
	 * Скорее всего так и происходит, но однажды почему-то не произошло.
	 */
	if (!get_permalink($post_id)) {
		delete_post_meta($post_id, $meta_key, $value);
		$post_id = null;
	}

	return $post_id;
}

/**
 * Получить id поста WordPress по названию продукта
 *
 * Используется для поиска существующих непривязанных товаров
 *
 * @param string $name Название продукта
 *
 * @return int Идентификатор поста
 */
function st_getPostIdByProductName($name)
{
	global $wpdb;
	$query = $wpdb->prepare(
		'SELECT ID FROM ' . $wpdb->posts . " WHERE post_type = 'product' AND post_title = %s",
		$name
	);
	$data = $wpdb->get_row($query, ARRAY_A);
	if (isset($data) && isset($data['ID'])) {
		return intval($data['ID']);
	}
}

/**
 * Получить id поста WordPress по названию вариации продукта
 *
 * @param string $name Название вариации продукта
 *
 * @return int Идентификатор поста
 */
function st_getPostIdByProductVariationName($name)
{
	global $wpdb;
	if (preg_match('/^(.+) \\((.+)\\)$/', $name, $matches)) {
		$log_message = 'st_getPostIdByProductVariationName $name to ' . $matches[1] . ' - ' . $matches[2];
		st_log_debug($log_message);
		$name = $matches[1] . ' - ' . $matches[2];
	}
	$query = $wpdb->prepare(
		'SELECT ID FROM ' . $wpdb->posts . " WHERE post_type = 'product_variation' AND post_title = %s",
		$name
	);
	$data = $wpdb->get_row($query, ARRAY_A);
	if (isset($data) && isset($data['ID'])) {
		return intval($data['ID']);
	}
}

/**
 * Получить цену товара или модификации по указанному прайсу
 *
 * @param array $good_data Данные продукта
 * @param int $price_id Идентификатор прайса
 *
 * @return float Цена продукта по выбранному прайсу
 */
function st_getProductPriceValue($good_data, $price_id)
{
	$price_value = null;
	foreach ($good_data['prices'] as $price_data) {
		if ($price_data['price_id'] == $price_id && $price_data['value'] != '') {
			$price_value = $price_data['value'];
			break;
		}
		// Цена в прайсе по умолчанию
		if ($price_data['is_required'] == true) {
			$price_default_value = $price_data['value'];
		}
	}

	// Если цены в прайсе нет - ставим цену по умолчанию
	if ($price_value == '') {
		$price_value = $price_default_value;
	}

	return floatval($price_value);
}

/**
 * Проверка что продукт в одной из импортируемых групп
 *
 * @param array $good Данные продукта
 * @param array $import_tags Группы для импорта
 *
 * @return bool true если пересечения есть, false если нет
 */
function st_checkProductInTags(array $good, array $import_tags): bool
{
	// Если не выбрана ни одна группа считаем что теги совпадают
	if ($import_tags == null || !is_array($import_tags) || count($import_tags) == 0) {
		return true;
	}

	// Если у товара нет тегов, то считаем что он не попадает
	if (!is_array($good['tags']) || count($good['tags']) == 0) {
		return false;
	}

	// Проверяем что теги пересекаются
	if (count(array_intersect($good['tags'], $import_tags)) > 0) {
		return true;
	}

	return false;
}

/**
 * Приведение значения к булевому типу.
 */
function force_bool($var)
{
	return boolval($var);
}

/**
 * Попытка привести значение к целому.
 * Если неудача - null.
 */
function force_int($var)
{
	if (is_numeric($var)) {
		return intval($var);
	}
	return null;
}

/**
 * Фильтрует массив (не ассоциативный), оставляя в нём только те элементы,
 * которые являются целыми числами. При этом производится попытка кастинга к целому.
 */
function force_int_keys($arr)
{
	/**
	 * Эта процедура используется в в хуке при чтении опции option_subtotal_import_stores
	 * Но пока не реализована функциональность выборки остатков
	 * по нескольким складам, склады хранятся как массив с одним элементом.
	 * tag: multi-stores
	 */
	$result = [];
	if (!is_array($arr)) {
		return $result;
	}

	$stored_value = array_map('intval', array_filter($arr, 'is_numeric'));
	if (count($stored_value)) {
		$result[] = $stored_value[0];
	}

	return $result;
}

/**
 * На входе json-строка, на выходе ассоциативный массив
 */
function json_assoc(string $var)
{
	return json_decode($var, true);
}

/**
 * На входе json-строка, на выходе массив
 */
function json_array($var)
{
	// Что-то не понятно как вызываются фильтры
	if (is_array($var)) {
		return $var;
	}

	$result = json_decode($var, true);
	if (!is_array($result)) {
		$result = [];
	}
	return $result;
}

/**
 * На входе строка с URL, на выходе строка с URL с единственным завершающим '/'
 */
function check_subtotal_url(string $var): string {
	return rtrim($var, '/') . '/';
}
