<?php
/**
 * Plugin Name: Subtotal
 * Plugin URI: https://subtotal.ru/support/integraczii/integracziya-s-internet-magazinom-woocommerce
 * Description: Subtotal integration with WooCommerce
 * Author: Subtotal
 * Author URI: https://subtotal.ru
 */
require_once 'functions.php';

/**
 * Subtotal API URL
 */
const SUBTOTAL_API_URL_DEFAULT = 'https://app.subtotal.ru/';

/**
 * Таймаут запросов к Subtotal в секундах
 */
const SUBTOTAL_REQUEST_TIMEOUT = 600;
/**
 * Класс-ответ из процедур, которые помимо статуса могут вернуть сообщение.
 */
class NamedStatus
{
	/**
	 * Как с unix'овым shell'ом - 0, значит ок. Другое - не ок.
	 * @type int (php 7.4)
	 */
	public $status;
	/**
	 * @type ?string (php 7.4)
	 */
	public $message;
}

/**
 * Структура-ответ из Subtotal API
 */
class StResponse extends NamedStatus
{
	/**
	 * Сырой ответ.
	 * @type array|WP_Error (php8)
	 */
	public $response;
	/**
	 * Любые данные, в зависимости от контекста вызова.
	 */
	public $data;
}

/**
 * Разделитель тегов.
 */
const TAG_SEPARATOR = '$%^|],';

/**
 * Время жизни блокировки в минутах
 */
const BLOCK_TTL = 30;

/**
 * Время обновления по расписанию в минутах
 * По умолчанию - сутки
 */
const CRON_UPDATE_PERIOD = 60 * 60 * 1

/**
 * Аутентификация пользователя.
 */
function st_Authentication(
	string $subtotal_email,
	string $subtotal_password
): StResponse {
	$result = new StResponse();

	if (!$subtotal_email || !$subtotal_password) {
		$result->status = 1;
		$result->message = '<div class="error">Неправильный логин/пароль</div>';
	}

	$SUBTOTAL_API = get_option('subtotal_api_url');
	$api_url = $SUBTOTAL_API . 'webapi/login';
	$request_args = [
		'body' => json_encode([
			'login' => $subtotal_email,
			'password' => $subtotal_password,
		]),
		'timeout' => SUBTOTAL_REQUEST_TIMEOUT
	];
	$response = wp_remote_post($api_url, $request_args);

	$result->response = $response;
	$session = null;
	if ($response instanceof WP_Error) {
		$result->status = 1;
		$result->message = "<div class='error'>Ошибка при обращении к Subtotal. Попробуйте ещё раз через некоторое время.\n";
		$result->message .= 'Если ошибка будет повторяться, сообщите данные в техподдержку: <br>';
		$result->message .= '<pre>Code: ' . ($response->get_error_code()) . "\nMessage: " . ($response->get_error_message()) . '</pre></div>';
	} else {
		// Reading cookie 'session'
		$result->status = 0;
		foreach ($response['cookies'] as $cookie) {
			if ($cookie->name == 'session') {
				$session = $cookie->value;
			}
		}
	}

	$result->data = $session;
	return $result;
}

/**
 * Получить список магазинов для данного бизнеса
 *
 * @return array Список магазинов  [id => имя, ...]
 */
function st_getPoses(): array
{
	$poses = [];
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		// Недокументированный метод API Subtotal
		// В официальном API нет способа получить список магазинов

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/poslist.php'
					. '?show_archive=false';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$poses_raw = json_decode(wp_remote_retrieve_body($request), true);
		if (!isset($poses_raw['poses'])) {
			// TODO: дампить ответ сервера в лог
			throw new Exception('Неверный ответ сервера на запрос списка точек продаж');
		}

		foreach ($poses_raw['poses'] as $pos) {
			$pos_id = $pos['id'];
			$pos_name = $pos['name'];
			$poses[$pos_id] = $pos_name;
		}
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage()
		];
	}
	return $poses;
}

/**
 * Получить список складов для данного бизнеса
 *
 * @return array Список складов [id => имя, ...]
 */
function st_getStores(): array
{
	$stores = [];
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		// Недокументированный метод API Subtotal
		// В официальном API нет способа получить список складов

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/storelist.php'
					. '?show_archive=false';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$stores_raw = json_decode(wp_remote_retrieve_body($request), true);
		if (!isset($stores_raw['stores'])) {
			// TODO: дампить ответ сервера в лог
			throw new Exception('Неверный ответ сервера на запрос списка складов');
		}

		foreach ($stores_raw['stores'] as $store) {
			$store_id = $store['id'];
			$store_name = $store['name'];
			$stores[$store_id] = $store_name;
		}
	} catch (Exception $e) {
		return [
			// TODO (мерж не принимать): проверить, что структура понимается JS.
			// Плохой подход: требуется постоянно щупать на наличие поля status
			'status' => false,
			'error' => $e->getMessage()
		];
	}
	return $stores;
}

/**
 * Получить каталог товаров (весь)
 *
 * @return array Каталог товаров
 */
function st_getCatalog()
{
	// Получение списка товаров
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/webapi/goods';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$catalog = json_decode(wp_remote_retrieve_body($request), true);
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
	return $catalog;
}

/**
 * Получить выгрузку товаров в указанном складе с указанным прайсом
 * в указанных количествах
 *
 * В отличие от getCatalog возвращает остатки и не разделяет товары
 * и модификации.
 *
 * @param int $store_id Склад
 * @param int $price_id Прайс
 * @param int $tags Прайс
 * @param int $page_num Страница (для постраничного запроса)
 * @param int $page_size Размер страницы (для постраничного запроса)
 * @param int $ch_from Дата последнего изменения
 *
 * @return array Список товаров включая модификации и объём на указанном складе
 */
function st_getStoreGoods(
	int $store_id,
	int $price_id,
	array $tags,
	int $page_num,
	int $page_size,
	$ch_from = 0
) {
	// Получение списка товаров
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$required_good_types = [
			GOOD_TYPE_ID_COMPOSED,
			GOOD_TYPE_ID_COMPOSED_CASHBOX,
			GOOD_TYPE_ID_GOODS
		];

		$required_tags = '';
		if (count($tags)) {
			$required_tags = implode(TAG_SEPARATOR, $tags);
		}

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/data/store_goods.php'
					. '?favourite_goods=0'
					. '&good_type=' . implode(',', $required_good_types)
					. '&page_num=' . $page_num
					. '&page_size=' . $page_size
					. '&price=' . $price_id
					. '&return_suggestions=false'
					. '&search='
					. '&store=' . $store_id
					. '&tags=' . $required_tags
					. '&totals=0'
					. '&no_store_sales=force_on'
					. '&ch_from=' . $ch_from;
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$catalog = json_decode(wp_remote_retrieve_body($request), true);
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
	return $catalog;
}

/**
 * Получить прайсы для текущего бизнеса
 *
 * @return array Список прайсов
 */
function st_getPriceLists()
{
	$price_lists = [];
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/repricing/get_prices.php'
					. '?is_service=false'
					. '&show_archive=false';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$prices_list_raw = json_decode(wp_remote_retrieve_body($request), true);
		if (!isset($prices_list_raw['repricing_prices'])) {
			// TODO: дампить ответ сервера в лог
			throw new Exception('Неверный ответ сервера на запрос списка прайслистов');
		}

		foreach ($prices_list_raw['repricing_prices'] as $price_list) {
			$price_list_id = $price_list['id'];
			$price_list_name = $price_list['full_name'];
			$price_lists[$price_list_id] = $price_list_name;
		}
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
	return $price_lists;
}

/**
 * Получить детальную информацию о товаре
 *
 * Если указать id модификации, вернёт родительский товар (и модификации к нему).
 * Так же нужна ради тегов.
 *
 * @param int $good_id Идентификатор товара
 */
function st_getGoodDetails($good_id)
{
	// Получение списка товаров
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/webapi/goods/' . $good_id;
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$details = json_decode(wp_remote_retrieve_body($request), true);
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
	return $details;
}

/**
 * Получить группы товаров для данного бизнеса (в WooCommerce это теги)
 *
 * @return array Список тегов
 */
function st_getGoodTags()
{
	// Получение списка групп
	// Недокументированный метод API
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/webapi/good_tags';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$good_tags = json_decode(wp_remote_retrieve_body($request), true);
		if (!isset($good_tags['tags'])) {
			// TODO: дампить ответ сервера в лог
			throw new Exception('Неверный ответ сервера на запрос списка групп (тегов)');
		}

		return $good_tags['tags'];
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}

	return $good_tags;
}

/**
 * Получить картинку товара
 *
 * Нужна потому что WordPress не умеет обращаться для загрузки картинки
 * с пользовательскими куками.
 *
 * @param string $image_url URL изображения в subtotal
 * @return mixed содержимое изображения, двоичные данные
 */
function st_getImage($image_url)
{
	$result = [
		'name' => '',
		'data' => null,
	];

	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$image_name = getImageName($image_url);
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($image_url, $request_args);
		st_throw_if_error($request);

		if ($request['response']['code'] == 404) {
			$message = 'Not found: ' . $image_url;
			st_log_info($message);
			return;
		}

		$result['data'] = wp_remote_retrieve_body($request);
		$result['name'] = $image_name . '.' . getImageExtension($result['data']);
	} catch (Exception $e) {
		throw new Exception('refactore me');
		// return [
		// 	'status' => false,
		// 	'error' => $e->getMessage(),
		// 	'data' => []
		// ];
	}

	return $result;
}

/**
 * Узнать тип файла с графикой.
 * Безусловный велосипед, но надо было быстро и хоть как-то.
 * Тут файл открывается, пишется, потом читается. In-memory аналога exif_imagetype
 * не нашёл. TODO: performance
 */
function getImageExtension($binary_data)
{
	$rand_str = bin2hex(random_bytes(10));
	$tmpfname = tempnam(sys_get_temp_dir(), $rand_str);
	$handle = fopen($tmpfname, 'wb');
	fwrite($handle, $binary_data);
	fclose($handle);

	$image_type = exif_imagetype($tmpfname);

	unlink($tmpfname);

	switch ($image_type) {
		case IMAGETYPE_GIF:
			$extension = 'gif';
			break;
		case IMAGETYPE_JPEG:
			$extension = 'jpeg';
			break;
		case IMAGETYPE_PNG:
			$extension = 'png';
			break;
		case IMAGETYPE_SWF:
			$extension = 'swf';
			break;
		case IMAGETYPE_PSD:
			$extension = 'psd';
			break;
		case IMAGETYPE_BMP:
			$extension = 'bmp';
			break;
		case IMAGETYPE_TIFF_II:
		case IMAGETYPE_TIFF_MM:
			$extension = 'tiff';
			break;
		case IMAGETYPE_JPC:
			$extension = 'jpc';
			break;
		case IMAGETYPE_JP2:
			$extension = 'jp2';
			break;
		case IMAGETYPE_JPX:
			$extension = 'jpx';
			break;
		case IMAGETYPE_JB2:
			$extension = 'jb2';
			break;
		case IMAGETYPE_SWC:
			$extension = 'swc';
			break;
		case IMAGETYPE_IFF:
			$extension = 'iff';
			break;
		case IMAGETYPE_WBMP:
			$extension = 'wbmp';
			break;
		case IMAGETYPE_XBM:
			$extension = 'xbm';
			break;
		case IMAGETYPE_ICO:
			$extension = 'ico';
			break;
		case IMAGETYPE_WEBP:
			$extension = 'webp';
			break;
		default:
			$extension = '';
			break;
	}

	return $extension;
}

/**
 * Выдергивание идентификатора изображения из URL вида
 * http://localhost:8081/id325/good29880_picviewimage1.php?uniq=-7345123875126185118
 */
function getImageName(string $image_url): string
{
	// Можно в одну строчку, а в несколько - нагляднее
	$parsed = parse_url($image_url);
	$query_as_array = explode('=', $parsed['query']);
	try {
		// Идентификатор отрицательное число. Пусть будет положительным ;)
		$result = (string) abs($query_as_array[1]);
	} catch (Exception $e) {
		// Если идентификатор не число
		$result = $query_as_array[1];
	}

	return $result;
}

/**
 * Экспорт продажи в Subtotal
 * Создаёт продажу с указанными параметрами в Subtotal
 *
 * @param array $sale Данные о продаже
 * @param integer $id Идентификатор существующей продажи
 * @return ?
 */
function st_exportSale($sale, $id = '')
{
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API . $owner_url . '/webapi/stock/outcomes' . ($id == '' ? '' : '/$id');
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'headers' => [
				'Content-Type' => 'application/json'
			],
			'body' => json_encode($sale),
			'method' => $id == '' ? 'POST' : 'PUT',
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_request($api_url, $request_args);
		st_throw_if_error($request);

		$response = json_decode(wp_remote_retrieve_body($request), true);
		return $response;
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
}

/**
 * Получение данных продажи
 *
 * @param integer $id Идентификатор продажи
 * @return array Продажа
 */
function st_getSale($id)
{
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. "/webapi/stock/outcomes/$id";
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$response = json_decode(wp_remote_retrieve_body($request), true);
		return $response;
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
}

/**
 * Поиск среди клиентов
 *
 * @param string $filter Строка для поиска
 * @return array Список подходящих клиентов
 */
function st_filterPartners($filter): StResponse
{
	$response = new StResponse();
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/partner/list.php'
					. '?filter_text=' . $filter;
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$response->status = 0;
		$partners_raw = json_decode(wp_remote_retrieve_body($request), true);
		$response->data = $partners_raw['partners'];
		return $response;
	} catch (Exception $e) {
		$response->status = 1;
		$response->message = $e->getMessage();
	}
	return $response;
}

/**
 * Получить список бизнесов у текущего аккаунта
 *
 * @return array Список бизнесов
 */
function st_getShops(): StResponse
{
	$response = new StResponse();
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. 'subtotal_common/dktapp/shop_list';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_get($api_url, $request_args);
		st_throw_if_error($request);

		$response->status = 0;
		$response->data = json_decode(wp_remote_retrieve_body($request), true);
	} catch (Exception $e) {
		$response->status = 1;
		$response->message = $e->getMessage();
	}
	return $response;
}

/**
 * Создание клиента (партнёра, покупателя)
 *
 * @param array $partnerData Данные клиента
 * @return array ответ от Subtotal
 */
function st_createPartner($partnerData)
{
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/partner/editnull.php';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'headers' => [
				'Content-Type' => 'application/json'
			],
			'body' => json_encode($partnerData),
			'method' => 'POST',
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_request($api_url, $request_args);
		st_throw_if_error($request);

		$response = json_decode(wp_remote_retrieve_body($request), true);
		return $response;
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
}

/**
 * Отметить заказ как оплаченный
 *
 * @param integer $order_id Номер заказа
 * @param array $payData Данные об оплате
 */
function st_payOrder($order_id, $payData)
{
	try {
		$access = get_access_info();
		$owner_url = $access['owner_url'];
		$session = $access['session'];

		$SUBTOTAL_API = get_option('subtotal_api_url');
		$api_url = $SUBTOTAL_API
					. $owner_url
					. '/store/outcomepay' . $order_id . '.php';
		$request_args = [
			'cookies' => [
				'session' => $session
			],
			'headers' => [
				'Content-Type' => 'application/json'
			],
			'body' => json_encode($payData),
			'method' => 'POST',
			'timeout' => SUBTOTAL_REQUEST_TIMEOUT
		];

		$request = wp_remote_request($api_url, $request_args);
		st_throw_if_error($request);

		$response = json_decode(wp_remote_retrieve_body($request), true);
		return $response;
	} catch (Exception $e) {
		return [
			'status' => false,
			'error' => $e->getMessage(),
			'data' => []
		];
	}
}

/**
 * Проверка авторизации. В случае успеха возвращается идентификатор сессии.
 * Иначе бросается исключение.
 */
function get_access_info(): array
{
	$access = get_option('subtotal_access');
	/**********  костыль для разработки  РЕВЬЮВЕР - УДОЛИТЬ!!!!  *****/
	$access['owner_url'] = 'id' . $access['owner_id'];
	if (!isset($access['session'])) {
		throw new Exception('Необходима авторизация');
	}

	return $access;
}
/**
 * Отладочная процедура, отправляет лог в syslog.
 *
 * @param string $message Сообщение в лог
 * @param string $severity Уровень логирования
 */
function st_log($message, $severity)
{
	syslog($severity, '[Subtotal] ' . $message);
	/**
	 * TODO: добавить Sentry
	 */
}

/**
 * Записать в лог с уровнем 7 (debug)
 */
function st_log_debug($message)
{
	return st_log($message, 7);
}

/**
 * Записать в лог с уровнем 6 (info)
 */
function st_log_info($message)
{
	return st_log($message, 6);
}

/**
 * Записать в лог с уровнем 5 (notice)
 */
function st_log_notice($message)
{
	return st_log($message, 5);
}

/**
 * Проверка ответа Subtotal на ошибку.
 * Если ошибка - то выбрасывается исключение.
 */
function st_throw_if_error($request) {
	if (is_wp_error($request)) {
		throw new Exception($request->get_error_message());
	}

	if (
		isset($request['response'])
		&& isset($request['response']['code'])
		&& $request['response']['code'] == 500
	) {
		throw new Exception($request['body']);
	}
}