<?php
/**
 * Plugin Name: Subtotal
 * Plugin URI: https://subtotal.ru/support/integraczii/integracziya-s-internet-magazinom-woocommerce
 * Description: Subtotal integration with WooCommerce
 * Author: Subtotal
 * Author URI: https://subtotal.ru
 */
require_once 'subtotal-api.php';

/**
 * Глобальные переменные. Для кеширования.
 */
$SYNCHRONIZED_GOODS_IDS = [];
$PRODUCT_VARIATIONS = [];
$BASE_PRODUCTS = [];
$VARIATION_PRODUCTS = [];
$LOADED_PRODUCTS = [];

/**
 * Процедура импорта по расписанию
 * Устанавливает всякие переменные состояния и запускает импорт
 * Используется в хуке subtotal_cron_hook
 *
 * Родственная процедура - st_importProducts_callback.
 * Отличие в том, что st_importProducts_callback отправляет все данные
 * на фронт.
 */
function subtotal_cron_exec()
{
	// Блокировка внутренних процессов,
	// выполняющихся при добавлении/изменении записей
	wp_defer_term_counting(true);
	wp_defer_comment_counting(true);

	try {
		importProducts(true);
	} catch (Exception $e) {
		st_log_info($e->getMessage());
	}

	// Разблокировка внутренних процессов,
	// выполняющихся при добавлении/изменении записей
	wp_defer_term_counting(false);
	wp_defer_comment_counting(false);
	wp_die();
}

/**
 * Процедура импорта по нажатию кнопки
 * Используется в хуке wp_ajax_st_importProducts
 *
 * Родственная процедура - subtotal_cron_exec.
 * Отличие в том, что subtotal_cron_exec не отправляет данные на фронт,
 * а пишет логи.
 */
function st_importProducts_callback()
{
	/**
	 * В идеале, импорт должен запускаться только у одного клиента.
	 * Все остальные клиенты должны блокироваться.
	 * В этом случае требуется мониторить активность того, кто запустил импорт.
	 * Например, требуется синхронизировать 10.000 товаров.
	 * Есть offset и limit.
	 * В опции (Update_option) пишется структура:
	 * "subtotal_update" : {
	 *     "started" : timestamp,
	 *     "offset" : int,
	 *     "limit" : int,
	 *     "last" : [timestamp] - массив, в который пушится timestamp
	 *                            после успешной синхронизации очередного (offset * limit)
	 *                            блока данных. После чего offset = offset + 1
	 * }
	 *
	 * Если есть активность (subtotal_update[list] равномерно заполняется), то ждать.
	 * Если активности нет - то продолжить синхронизацию.
	 * И добавить cron-задачу, которая бы удаляла зависшую синхронизацию.
	 */

	// Блокировка внутренних процессов,
	// выполняющихся при добавлении/изменении записей
	wp_defer_term_counting(true);
	wp_defer_comment_counting(true);

	$data_out = null;
	$has_error = false;

	try {
		$ids = importProducts(false);
	} catch (Exception $e) {
		$data_out = $e->getMessage();
		st_log_info($data_out);
		$has_error = true;
	}

	// Разблокировка внутренних процессов,
	// выполняющихся при добавлении/изменении записей
	wp_defer_term_counting(false);
	wp_defer_comment_counting(false);

	// wp_die вызывается внутри wp_send_json_*
	if ($has_error) {
		wp_send_json_error($data_out, 400); // Всегда возвращается 400. Временная мера
	} else {
		wp_send_json_success(['count' => count($ids)]);
	}
}

/**
 * Возвращает количество тегов и товаров в списке
 * Используется в хуке wp_ajax_st_getCatalogCount
 */
function st_getCatalogCount_callback()
{
	$qs = ['count' => true];
	$result = null;
	if (isset($_POST['groups']) && $_POST['groups'] == true) {
		// Количество тегов
		$good_tags = st_getGoodTags();
		$result = ['count' => count($good_tags)];
	} else {
		// Единственный способ узнать количество товаров -
		// получить полный каталог товаров
		$catalog = st_getCatalog();
		$result = ['count' => $catalog['length']];
	}
	wp_send_json_success($result);
}

/**
 * Импортирует группы товаров из Subtotal в WooCommerce
 * Используется в хуке wp_ajax_st_importGoodTags
 */
function st_importGoodTags_callback()
{
	$tags = st_getGoodTags();
	update_option('subtotal_good_tags', json_encode($tags));
	$import_tags_list = get_option('subtotal_import_good_tags');

	foreach ($tags as $tag) {
		/**
		 * Пропустить группы товаров, которые не были указаны к синхронизации.
		 */
		if ($import_tags_list !== null &&
			!in_array($tag, $import_tags_list)) {
			continue;
		}

		/**
		 * Создать категорию, если он ещё не существует
		 * В wp_insert_term можно передать третий аргумент:
		 *     array(
		 *         'description' => 'Description for category', // optional
		 *         'parent' => 0, // optional
		 *         'slug' => 'my-new-category' // optional
		 *     )
		 */
		if (!term_exists($tag, 'product_cat')) {
			wp_insert_term($tag, 'product_cat');
		}
	}

	wp_send_json_success($tags);
}

/**
 * Собственно логика обновления товаров.
 *
 * @param bool $auto Признак импорта по расписанию
 */
function importProducts($auto = false): array
{
	global $SYNCHRONIZED_GOODS_IDS;

	// Отмена ограничений по памяти
	ini_set('memory_limit', -1);

	// Отмена ограничений времени выполнения, игнорирование сброса
	set_time_limit(0);
	ignore_user_abort(true);

	// Отмена требования уникальности артикула (sku)
	add_filter('wc_product_has_unique_sku', '__return_false');

	// TODO: проверить, что страница начинается с 1, а не с 0
	if (isset($_POST['page_num'])) {
		$page_num = intval($_POST['page_num']);
	} else {
		$page_num = 1;
	}

	if ($page_num < 0) {
		$page_num = 1;
	}

	// Текущие настройки Subtotal
	$stores = (array) get_option('subtotal_import_stores');
	$pos_id = get_option('subtotal_import_pos');
	$tags = get_option('subtotal_import_good_tags');
	$price_id = get_option('subtotal_import_price');
	$page_size = get_option('subtotal_import_limit_value');

	// TODO: проверку настроек вынести в отдельную процедуру
	if (
		!count($stores)
		|| !$pos_id
		|| !$price_id
		|| !$page_size
	) {
		throw new Exception('Отсутствуют обязательные настройки.');
	}

	/**
	 * TODO: пока не реализована функциональность выборки остатков
	 * по нескольким складам, склады хранятся как массив с одним элементом.
	 * tag: multi-stores
	 */
	$store_id = $stores[0];

	// Кастинг к int на случай null
	$last_import_timestamp = (int)get_option('subtotal_last_import_timestamp');
	$start_time = time();

	$catalog = st_getStoreGoods(
		$store_id,
		$price_id,
		$tags,
		$page_num,
		$page_size, // Оно!!!
		$last_import_timestamp
	);

	if (isset($catalog['status']) && $catalog['status']) {
		$log_message = '';
		if (isset($catalog['error'])) {
			$log_message = $catalog['error'];
		}
		else {
			$log_message = 'Ошибка получения товаров из Subtotal';
		}
		throw new Exception($log_message);
	}

	/**
	 * Логика частичной синхронизации проста:
	 * 1. Если опция "uncomplete_synchronization_goods_ids" существует
	 *    и не пустой массив, то в ней хранятся id синхронизированных товаров.
	 * 2. Если id товара в "uncomplete_synchronization_goods_ids",
	 *    то товар пропускается.
	 * 3. Если синхронизация товаров прошла успешно,
	 *    то выставляется опция "subtotal_last_import_timestamp".
	 * 4. Если опция "subtotal_last_import_timestamp" не 0,
	 *    то при выборке используется товаров используется ch_from
	 */
	$uncomplete_synchronization_goods_ids = [];
	if (!$last_import_timestamp) {
		$uncomplete_synchronization_goods_ids = get_option('uncomplete_synchronization_goods_ids');
		if (!is_array($uncomplete_synchronization_goods_ids)) {
			$uncomplete_synchronization_goods_ids = [];
		}
	}

	if (isset($catalog['goods'])) {
		foreach ($catalog['goods'] as $good) {
			/**
			 * TODO (на следующий рефакторинг):
			 * недостаточно просто возвращать true/false.
			 * Если false, то требуется вернуть id, наименование, причину неудачи.
			 */
			$subtotal_good_id = $good['id'];
			if (!in_array($subtotal_good_id, $uncomplete_synchronization_goods_ids)) {
				st_insertOrUpdateProduct($good, $auto);
				$uncomplete_synchronization_goods_ids[] = $subtotal_good_id;
				// небольшой оверхед на сохранение синхронизированных товаров
				update_option('uncomplete_synchronization_goods_ids', $uncomplete_synchronization_goods_ids);
			}
		}
	}

	// Обновляем параметр времени последней синхронизации
	// Если замерить, например, $end_time, то можно посчитать время синхронизации, кеп!
	update_option('subtotal_last_import_timestamp', $start_time);
	update_option('uncomplete_synchronization_goods_ids', []);
	// Вот тут вот надо вернуть причину неудачи из TODO выше
	return $SYNCHRONIZED_GOODS_IDS;
}
